/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001-2002 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#include "nv-misc.h"
#include "os-interface.h"
#include "nv.h"
#include "nv-freebsd.h"

static d_open_t  nvidia_dev_open;
static d_close_t nvidia_dev_close;
static d_ioctl_t nvidia_dev_ioctl;
static d_poll_t  nvidia_dev_poll;
static d_mmap_t  nvidia_dev_mmap;

static struct cdevsw nvidia_dev_cdevsw = {
    .d_open =      nvidia_dev_open,
    .d_close =     nvidia_dev_close,
    .d_ioctl =     nvidia_dev_ioctl,
    .d_poll =      nvidia_dev_poll,
    .d_mmap =      nvidia_dev_mmap,
    .d_name =      "nvidia",
    .d_version =   D_VERSION,
    .d_flags =     D_MEM|D_TRACKCLOSE|D_NEEDGIANT
};

struct clonedevs *nvidia_dev_clones;

void nvidia_dev_clone(
    void *arg,
#if __FreeBSD_version >= 600034
    struct ucred *cred,
#endif
    char *name,
    int namelen,
    struct cdev **dev
)
{
    int i, unit;
    struct nvidia_softc *sc;
    int clone = 0;

    if (*dev != NULL)
        return;
    if (dev_stdclone(name, NULL, "nvidia", &unit) != 1)
        return;
    if (unit >= NV_MAX_DEVICES)
        return;

    sc = devclass_get_softc(nvidia_devclass, unit);
    if (sc == NULL)
        return;

    /*
     * XXX clone_create() doesn't handle automatic clone
     * number allocation correctly at this point, it
     * can insert clone list entries out of order, with
     * respect to the clone number, and subsequently
     * allocate numbers that are still in use. Iterate over
     * the possible clone numbers until the first unused
     * one is found.
     */
    do {
        i = clone_create(&nvidia_dev_clones, &nvidia_dev_cdevsw,
                &clone, dev, 0);
        if (i == 0) clone++;
    } while ((clone <= CLONE_UNITMASK) && (i == 0));

    if ((i != 0) && (clone <= CLONE_UNITMASK)) {
        *dev = make_dev(&nvidia_dev_cdevsw, unit2minor(clone),
                    UID_ROOT, GID_WHEEL, 0666, "nvidia%u.%u", unit, clone);
        if (*dev != NULL) {
            (*dev)->si_drv1 = (void *)sc;
            (*dev)->si_flags |= SI_CHEAPCLONE;
        }
    }
}

int nvidia_dev_open(
    struct cdev *dev,
    int oflags,
    int devtype,
    d_thread_t *td
)
{
    int status;
    struct nvidia_softc *sc = NV_GET_SOFTC(dev);
    nv_state_t *nv = sc->nv_state;

    struct nvidia_drv2 *drv2;

    if (!dev->si_drv2) {
        drv2 = malloc(sizeof(struct nvidia_drv2), M_NVIDIA, M_WAITOK | M_ZERO);
        if (drv2 == NULL)
            return ENOMEM;
        STAILQ_INIT(&drv2->event_queue);
        dev->si_drv2 = drv2;
    } else
        return EBUSY;

    nv_lock_api(nv);
    status = nvidia_open_dev(sc);
    nv_unlock_api(nv);

    if (status == 0) {
        /* XXX Fix me? (clear of SI_CHEAPCLONE) */
        dev->si_flags &= ~SI_CHEAPCLONE;
    } else
        free(drv2, M_NVIDIA);

    return status;
}

int nvidia_dev_close(
    struct cdev *dev,
    int fflag,
    int devtype,
    d_thread_t *td
)
{
    int status;
    struct nvidia_softc *sc = NV_GET_SOFTC(dev);
    struct nvidia_drv2 *drv2;

    nv_state_t *nv = sc->nv_state;
    struct nvidia_event *et;

    nv_lock_api(nv);
    status = nvidia_close_dev(sc, dev, td);
    nv_unlock_api(nv);

    drv2 = dev->si_drv2;
    while ((et = STAILQ_FIRST(&drv2->event_queue))) {
        STAILQ_REMOVE(&drv2->event_queue, et, nvidia_event, queue);
        free(et, M_NVIDIA);
    }

    if (status == 0) {
        free(dev->si_drv2, M_NVIDIA);
        /* XXX Fix me? (call to destroy_dev()) */
        destroy_dev(dev);
    }

    return status;
}

int nvidia_dev_ioctl(
    struct cdev *dev,
    u_long cmd,
    caddr_t data,
    int fflag,
    d_thread_t *td
)
{
    int status;
    struct nvidia_softc *sc = NV_GET_SOFTC(dev);
    nv_state_t *nv = sc->nv_state;

    if (__NV_IOC_TYPE(cmd) != NV_IOCTL_MAGIC)
        return ENOTTY;

    nv_lock_api(nv);
    status = nvidia_handle_ioctl(dev, cmd, data, fflag, td);
    nv_unlock_api(nv);

    return status;
}

int nvidia_dev_poll(
    struct cdev *dev,
    int events,
    d_thread_t *td
)
{
    struct nvidia_softc *sc = NV_GET_SOFTC(dev);
    nv_state_t *nv = sc->nv_state;

    struct nvidia_drv2 *drv2 = dev->si_drv2;

    nv_lock_rm(nv);

    if (STAILQ_EMPTY(&drv2->event_queue)) {
        nv_unlock_rm(nv);
        selrecord(td, &drv2->rsel);
    } else {
        nv_unlock_rm(nv);
        return (events & (POLLIN | POLLPRI));
    }

    return 0;
}

int nvidia_dev_mmap(
    struct cdev *dev,
    vm_offset_t offset,
    vm_offset_t *address,
    int nprot
)
{
    int status;
    struct nvidia_softc *sc = NV_GET_SOFTC(dev);
    nv_state_t *nv = sc->nv_state;

    vm_offset_t physical;

    nv_lock_api(nv);
    status = nvidia_mmap_dev(sc, offset, &physical);
    nv_unlock_api(nv);

    if (status != -1)
        *address = physical;

    return status;
}

/* XXX This needs to be populated with FreeBSD specific code to allow
 * us to support DMA addresses larger than 32 bits.
 */

void nv_set_dma_address_size(
    nv_state_t  *nv,
    U032        phys_addr_bits
)
{
}
